// ====================================================================
//  Class:  OLStats.OLGameStats
//  Parent: IpDrv.MasterServerGameStats
//
//  the GameStats object is used to send individual stat events to the
//  stats server.  Each game should spawn a GameStats object if it
//  wishes to have stat logging.
//
// ====================================================================

class OLLocalGameStats extends GameStats
    config(OLStats);

var array<OLStatsAccuracyCounter> AccuracyCounters; // Each player has one of these to track his accuracy
var array<int> BotArray;            // An array of all the bot playernums. Used to determine when a new bot has joined
var array<int> PlayerArray;         // An array of all the player playernums. Used to determine if someone gets into the game without a connect event.
var bool bExportRawLog;
var bool bExportHTMLLogs;
var bool bLogBots;              // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var bool bLogChat;              // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var bool bLogPickups;           // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var bool bLogVehicleDeaths;     // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var bool bLogSecurityInfo;      // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var bool bLogPings;             // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var bool bLogAccuracy;          // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var bool bAdvancedVoteLogging;  // OLStats options are controlled in MutOLGameStats, and these vars get updated from there.
var config bool bRunUTStatsDBParser;
var config string ParserURL;
var config string ParserPass;
var config bool bUploadLogs;
var config string UploadURL;
var config string UploadPass;
var() string version;               // The current version of OLStats. Updated for each new release.
var float TimeSinceLastPingReport;  // The time passed since pings were last reported.
var() int PingReportInterval;       // How often to report pings.

// Cache vars
var int InvasionWaveCache;          // The current Invasion wave in Invasion games.
var controller MutantCache;         // The current mutant in mutant games.
var controller BottomFeederCache;   // The current bottom feeder in mutant games.
var TeamInfo DomPointTeamCache[2];  // The current team that controls each point

var bool bEndGameExecuted;  // Set to true once the EndGame() function is called.
var bool bGameStatsJustRemoved; // Set to true when EndGame() is called. The very next tick it is set back to false.
var bool bExportedHTMLLogs; // Set to true after the HTML logger is told to export its logs.
var string EndGameReason; // The reason the game ended. Used when the HTML logging is exported one tick after game end.

var config bool bDebug;         // Debug mode makes all stat lines appear in the system log, for easier debugging.
var config bool bBroadcastLog;  // Broadcasts all stat lines to all players for debugging; Requires debug mode to be on

// Helper class variables
var() string RPGHelperName;
var OLStatsHelpers RPGHelper;

// HTML exporter variables
var OLStatsHTMLLogger EGS;

var HttpSock LogParserSock, SendLogSock;

var string TheLogName;
var array<string> LogArray;

var class<Invasion> UnusedInvasion; // Simple hack to make UDE realize that Invasion is a required package.
var class<ONSMobileAssaultStation> UnusedMAS; // Same as above
var class<UTServerAdminSpectator> UnusedSpec; // Same as above

// HTML Specific config variables;
var string DateTimeFormat;
var string DateFormat;

var int MinPlayers;
var int MinHumans;
var bool KeepBotRecords;


function Logf(string LogString)
{
    if (bDebug)
    {
        Log(LogString,'OLStats');
        if (bBroadcastLog)
            Level.Game.Broadcast(self, "OLStats:"@repl(LogString,Tab," "));
    }

    if (TempLog!=None)
        TempLog.Logf(LogString);

    LogArray[LogArray.Length] = LogString;
}

/////////////////////////////////////
// GameStats interface

event PreBeginPlay()
{
    local Mutator Mut;
    local bool bMutAlreadySpawned;

    Super.PreBeginPlay();

    bMutAlreadySpawned = false;
    foreach AllActors(class'Mutator',Mut)
    {
        if(MutOLGameStats(Mut) != none)
        {
            bMutAlreadySpawned = true;
            break;
        }
    }

    if (!bMutAlreadySpawned)
        Level.Game.AddMutator("OLStats.MutOLGameStats");
}

function Init()
{
    Log("----------------------------------------",'OLStats');
    Log("OLStats Loading...",'OLStats');
    if (bDebug)
        Log("Debug mode is ON!",'OLStats');

    bExportRawLog = class'MutOLGameStats'.default.bExportRawLog;
    bExportHTMLLogs = class'MutOLGameStats'.default.bExportHTMLLogs;
    bLogBots = class'MutOLGameStats'.default.bLogBots;
    bLogChat = class'MutOLGameStats'.default.bLogChat;
    bLogPickups = class'MutOLGameStats'.default.bLogPickups;
    bLogVehicleDeaths = class'MutOLGameStats'.default.bLogVehicleDeaths;
    bLogSecurityInfo = class'MutOLGameStats'.default.bLogSecurityInfo;
    bLogPings = class'MutOLGameStats'.default.bLogPings;
    bAdvancedVoteLogging = class'MutOLGameStats'.default.bAdvancedVoteLogging;

    bRunUTStatsDBParser = class'MutOLGameStats'.default.bRunUTStatsDBParser;
    ParserURL = class'MutOLGameStats'.default.ParserURL;
    ParserPass = class'MutOLGameStats'.default.ParserPass;
    bUploadLogs = class'MutOLGameStats'.default.bUploadLogs;
    UploadURL = class'MutOLGameStats'.default.UploadURL;
    UploadPass = class'MutOLGameStats'.default.UploadPass;

    MinPlayers = class'MutOLGameStats'.default.MinPlayers;
    MinHumans = class'MutOLGameStats'.default.MinHumans;
    KeepBotRecords = class'MutOLGameStats'.default.KeepBotRecords;

    if ( Invasion(Level.Game) == none )
        bLogAccuracy = class'MutOLGameStats'.default.bLogAccuracy;
    else
        bLogAccuracy = false;

    TheLogName = GetLogFilename();
    if(bExportRawLog)
    {
        TempLog = spawn(class 'FileLog');
        if (TempLog != None)
        {
            TempLog.OpenLog(TheLogName);
        }
        else {
            Warn("Could not create output file");
        }
    }

    Log("OLStats "$version$" loaded!",'OLStats');
    Log("  Author: OverloadUT",'OLStats');
    Log("  Email: Overload@planetunreal.com",'OLStats');
    Log("  Web: www.apartment167.com",'OLStats');
    Log("----------------------------------------",'OLStats');

    if (bRunUTStatsDBParser)
    {
        RunUTStatsDBParser();
    }

    if (bExportHTMLLogs)
    {
        EGS = spawn(class'OLStatsHTMLLogger');
        if (EGS == none)
        {
            Warn("Failed to spawn OLStatsHTMLLogger");
            bExportHTMLLogs = false;
        } else {
            EGS.GS = self;
        }
    }
}

// Sends the log to a remote (or local) server at the end of the game.
function SendLogToServer()
{
    local int i;

    SendLogSock = spawn(class'LibHTTP2.HttpSock');
    if (SendLogSock == none)
    {
        Warn("Could not spawn LibHTTP2.HttpSock");
        return;
    }

    if (bDebug)
        SendLogSock.iVerbose = 10;

    SendLogSock.AddHeader("Content-Type", "application/x-www-form-urlencoded");
    SendLogSock.RequestData.Length = 1;
    SendLogSock.RequestData[0] = "pass="$UploadPass;
    SendLogSock.RequestData[0] $= "&filename="$TheLogName;
    SendLogSock.RequestData[0] $= "&data="$class'LibHTTP2.HttpUtil'.static.RawUrlEncode(LogArray[0])$Chr(13)$Chr(10);

    for(i=1;i<LogArray.Length;i++)
        SendLogSock.RequestData[0] $= class'LibHTTP2.HttpUtil'.static.RawUrlEncode(LogArray[i])$Chr(13)$Chr(10);

    SendLogSock.OnComplete = SendLockSockOnComplete;
    SendLogSock.HttpRequest(UploadURL,"POST");
}


// Do a GET on the UTStatsDB logs.php file, therefore executing the parser.
function RunUTStatsDBParser()
{
    LogParserSock = spawn(class'LibHTTP2.HttpSock');
    if (LogParserSock == none)
    {
        Warn("Could not spawn LibHTTP2.HttpSock");
        return;
    }

    if (bDebug)
        LogParserSock.iVerbose = 10;

    LogParserSock.OnComplete = ParserSockOnComplete;
    LogParserSock.HttpRequest(ParserURL$"?pass="$ParserPass);
}

function SendLockSockOnComplete()
{
    local int status, returncode, i;

    status = SendLogSock.LastStatus;

    Log("----------------------------------------",'OLStats');
    Log("OLStats Stat Uploader Operation Complete",'OLStats');

    if (status == 200)
    {
        if (bDebug)
        {
            Log("Return length: "$SendLogsock.ReturnData.Length);
            for (i=0;i<SendLogsock.ReturnData.Length;i++)
                Log("Row "$i$":"$SendLogsock.ReturnData[i]);
        }

        if ( SendLogSock.ReturnData[0] ~= "OLStatsSendLogTarget" )
        {
            if (SendLogSock.ReturnData.Length > 1)
                returncode = Int(SendLogSock.ReturnData[1]);
            else
                returncode = 998;
        }
        else
            returncode = 999;

        Log("  Return code: "$returncode,'OLStats');

        if (returncode == 1)
        {
            Log("  Upload successful!",'OLStats');
        } else if (returncode == 2) {
            Log("  ERROR: Incorrect password.",'OLStats');
            Log("  Check your UploadPass in OLStats.ini",'OLStats');
        } else if (returncode == 3) {
            Log("  ERROR: No filename specified. This should never happen.",'OLStats');
        } else if (returncode == 4) {
            Log("  ERROR: No data sent.  Empty log?",'OLStats');
        } else if (returncode == 999) {
            Log("  ERROR: Bad URL",'OLStats');
            Log("  Check to make sure you have the correct",'OLStats');
            Log("  URL specified in OLStats.ini",'OLStats');
        } else {
            Log("  ERROR: Unknown error.",'OLStats');
            Log("  Check to make sure you have the correct",'OLStats');
            Log("  URL specified in OLStats.ini",'OLStats');
            Log("  Also try manually entering the URL in a browser",'OLStats');
            Log("  and check for errors.",'OLStats');
        }
    } else {
        Log("  ERROR: Server responded:"@status@class'LibHTTP2.HttpUtil'.static.HTTPResponseCode(status),'OLStats');
    }

    Log("----------------------------------------",'OLStats');
}

function ParserSockOnComplete()
{
    local int status, i;
    local string CurString;
    local string LastString;

    status = LogParserSock.LastStatus;

    Log("----------------------------------------",'OLStats');
    Log("OLStats Remote Parser Operation Complete",'OLStats');

    if (status == 200)
    {
        Log("Parser Result:",'OLStats');

        LastString = LogParsersock.ReturnData[LogParserSock.ReturnData.Length-1];

        if( InStr(LogParserSock.ReturnData[0],"Access error.") > -1 )
        {
            Log("  Access Error.",'OLStats');
            Log("  Check your ParserPass in OLStats.ini",'OLStats');
        } else if( InStr(LastString,"logs processed") > -1 ) {
            CurString = LastString;
            Log("  "$Mid( CurString,3,Len(CurString)-15 ),'OLStats');
            for(i=1;i<LogParserSock.ReturnData.Length-1;i++)
            {
                CurString = LogParsersock.ReturnData[i];
                Log("  "$Mid(CurString,0,Len(CurString)-4) ,'OLStats');
            }
        } else if( InStr(LastString,"Calculating Career Highs") > -1 || InStr(LastString,"Finished") > -1 ) {
            CurString = LogParserSock.ReturnData[LogParserSock.ReturnData.Length-2];
            Log("  "$Mid( CurString,3,Len(CurString)-15 ),'OLStats');
            for(i=1;i<LogParserSock.ReturnData.Length-2;i++)
            {
                CurString = LogParsersock.ReturnData[i];
                Log("  "$Mid(CurString,0,Len(CurString)-4) ,'OLStats');
            }
        } else if( InStr(LastString,"No log files to process.") > -1 ) {
            Log("  No log file to process.",'OLStats');
            Log("",'OLStats');
        } else if( InStr(LastString,"Log parser is currently locked for update.") > -1 ) {
            Log("  Log parser is currently locked.",'OLStats');
            Log("",'OLStats');
        } else {
            Log("  Parser returned unknown data.",'OLStats');
            Log("  This does not mean that it was unsuccessful. Check your stats page to be sure.",'OLStats');
            Log("  Check to make sure the URL is correct in OLStats.ini",'OLStats');
            Log("  Also try running the parser manually and see if any strange errors are occuring.",'OLStats');
        }
    } else {
        Log("  ERROR: Server responded:"@status@class'LibHTTP2.HttpUtil'.static.HTTPResponseCode(status),'OLStats');
    }

    Log("----------------------------------------",'OLStats');
}

function Shutdown()
{
    if (!bEndGameExecuted)
        EndGame("Unknown");

    if (TempLog!=None)
        TempLog.Destroy();
}



/////////////////////////////////////
// Internals

function string TimeStamp()
{
    local string seconds;
    seconds = string(Level.TimeSeconds);

    // No longer removing centiseconds.

    // Remove the centiseconds
    // if( InStr(seconds,".") != -1 )
    //    seconds = Left( seconds, InStr(seconds,".") );

    return seconds;
}

function string Header()
{
    return TimeStamp()$Tab;
}

function String FullTimeDate()      // Date/Time in MYSQL format
{
    return Level.Year$"-"$Level.Month$"-"$Level.Day$" "$Level.Hour$":"$Level.Minute$":"$Level.Second;
}

function String TimeZone()          // Timezone (offset) of game server's local time to GTM, e.g. -4 or +5
{
    return "0";                     // FIXME Jack - currently pretending GMT
}

function String MapName()
{
    local string mapname;

    mapname = GetMapFileName();

    // Remove the file name extention .ut2
    if( InStr(mapname,".ut2") != -1 )
        mapname = Left( mapname, InStr(mapname,".ut2") );

    ReplaceText(mapname, tab, "_");

    return mapname;
}


// Stat Logging functions
function NewGame()
{
    local string out, tmp;
    local string ngTitle, ngAuthor, ngGameGameName;
    local int i;
    local mutator MyMutie;
    local GameRules MyRules;
    local OLGameStatRules GSR;
    local OLBroadcastHandler BH;
    local BroadcastHandler OldHandler;
    local OLStatsVotingHandler VH;
    local bool bHelperMutAlreadySpawned;

    local class<OLStatsHelpers> RPGHelperClass;

    bEndGameExecuted = false;

    Logf("OLStats v"$Version$" by OverloadUT www.Apartment167.com");

    ngTitle         = Level.Title;          // Making local copies
    ngAuthor        = Level.Author;
    ngGameGameName  = Level.Game.GameName;
    ReplaceText(ngTitle,        tab, "_");  // Replacing tabs with _
    ReplaceText(ngAuthor,       tab, "_");
    ReplaceText(ngGameGameName, tab, "_");

    GRI = Level.Game.GameReplicationInfo;
    out = Header()$"NG"$Tab;                // "NewGame"
    out $= FullTimeDate()$Tab;      // Game server's local time
    out $= TimeZone()$Tab;          // Game server's time zone (offset to GMT)
    out $= MapName()$Tab;               // Map file name without map extention .ut2
    out $= ngTitle$Tab;
    out $= ngAuthor$Tab;
    out $= Level.Game.Class$Tab;
    out $= ngGameGameName;

    tmp = "";
    i = 0;
    bHelperMutAlreadySpawned = false;
    foreach AllActors(class'Mutator',MyMutie)
    {
        if (tmp != "")
            tmp $= "|" $ MyMutie.Class;
        else
            tmp $= MyMutie.Class;

        i++;

        if ( MyMutie.IsA('MutUT2004RPG') && RPGHelper == none)
        {
            log("****************************************");
            log("OLStats Special Mutator Compatability");
            log("UT2004 RPG detected");
            log("Spawning special UT2004RPG Stats handler");
            RPGHelperClass = class<OLStatsHelpers>(DynamicLoadObject(RPGHelperName, class'Class'));
            RPGHelper = spawn(RPGHelperClass);
            if (RPGHelper != none)
                log("Spawn Successful!");
            else
                log("!WARNING! FAILED TO SPAWN RPG HELPER CLASS");
            log("****************************************");
        }

        if ( MutOLGameStats(MyMutie) != none )
        {
            bHelperMutAlreadySpawned = true;
        }
    }
    foreach AllActors(class'GameRules',MyRules)
    {
        if (tmp!="")
            tmp $= "|"$MyRules.Class;
        else
            tmp $= MyRules.Class;

        i++;
    }

    if (i>0)
    {
        ReplaceText(tmp, tab, "_");
        out $= Tab $ "Mutators=" $ tmp;
    }


    // I need to spawn my own GameRules in order to track pickups and other events
    GSR = spawn(class'OLGameStatRules');
    Level.Game.AddGameModifier(GSR);
    GSR.GS = self;

    // Now I replace the BroadcastHandler of the level with my own so that I can
    // track chat. I set the "real" BroadcastHandler as the oldHandler
    // of my BH that way it minimizes conflict with other mutators.
    OldHandler = Level.Game.BroadcastHandler;
    BH = spawn(class'OLBroadcastHandler');
    BH.GS = self;
    Level.Game.BroadcastHandlerClass = "OLStats.OLBroadcastHandler";
    Level.Game.BroadcastClass = class'OLBroadcastHandler';
    Level.Game.BroadcastHandler = BH;
    BH.oldHandler = oldHandler;

    // And I need to make my own VotingHandler so I can track voting.
    // First verify that there is a voting handler, and make sure that it is the
    // built in one; Do not replace one from a custom voting mod!
    if (bAdvancedVoteLogging && Level.Game.VotingHandler != none && Level.Game.VotingHandler.Class == class'xVoting.xVotingHandler')
    {
        VH = spawn(class'OLStatsVotingHandler');
        VH.GS = self;
        if (VH != none)
        {
            Level.Game.VotingHandler.Destroy();
            Level.Game.VotingHandlerType = "OLStatsVotingHandler";
            Level.Game.VotingHandlerClass = class'OLStatsVotingHandler';
            Level.Game.VotingHandler = VH;
        } else {
            log("WARNING: Failed to spawn the OLStats Voting Handler!");
        }
    }

    Logf(out);
    if (bExportHTMLLogs)
    {
        EGS.NewGame();
        EGS.MapName(MapName());
    }
}

function LogPickup(PlayerReplicationInfo Who, Pickup item)
{
    local string out;
    local string Classname, Packagename;
    local string TheClass;

    if (!bLogPickups)
        return;

    if ( Who.bOnlySpectator )           // Spectators should never show up in stats!
        return;

    // if it's a practice round, then we don't log it.
    if ( ASGameInfo(Level.Game) != none && ASGameInfo(Level.Game).IsPracticeRound() )
        return;

    if (Who.bBot)
    {
        if (!Checkbot(Who))
            return;
    } else
        CheckPlayer(Who);

    TheClass = string(item.class);

    if(RPGHelper != none)
        TheClass = RPGHelper.RPGGetRealPickup(item);

    Divide(TheClass,".",Packagename,Classname);
    out = Header()$"I"$Tab$Controller(Who.Owner).PlayerNum$Tab$Classname;

    Logf(out);

    if (bExportHTMLLogs)
        EGS.PickupItem(Who,Item);
}

function LogMapVote(PlayerReplicationInfo Who, coerce string MapName, int VotePoints)
{
    local string out;

    out = Header()$"MV"$Tab$Controller(Who.Owner).PlayerNum$Tab$MapName$Tab$VotePoints;

    logf(out);
}

function LogKickVote(PlayerReplicationInfo Who, int Victim)
{
    local string out;
    local int PlayerNum;

    if(Who == none)
        PlayerNum = -2;
    else
        PlayerNum = Controller(Who.Owner).PlayerNum;

    out = Header()$"KV"$Tab$PlayerNum$Tab$string(Victim);

    logf(out);
}

// Special chat event that comes from an xVotingHandler
function LogVoteChat(coerce string Msg)
{
    local string out;

    if (bAdvancedVoteLogging)
        return;

    out = Header()$"V"$Tab$"-1"$Tab$Msg;

    Logf(out);
}

// Any chat event that comes from an actor.
function LogChat(PlayerReplicationInfo Who, coerce string Msg, string ChatType)
{
    local string out;
    local int PlayerNum;

    if (!bLogChat)
        return;

    if (Who == none)
        return;

    if ( UTServerAdminSpectator(Who.Owner) == none )
    {
        if ( Who.bOnlySpectator )   // Spectators should never show up in stats!
            return;

        if ( Controller(Who.Owner) == none )
            return;

        if (Who.bBot)
        {
            if (!Checkbot(Who))
                return;
        } else
            CheckPlayer(Who);

        PlayerNum = Controller(Who.Owner).PlayerNum;
    } else {
        PlayerNum = -1;
    }

    out = Header()$ChatType$Tab$PlayerNum$Tab;
    out $= Msg;

    Logf(out);

//    if(bExportHTMLLogs)
//        EGS.LogMessage(Who, Msg, ChatType);
}

// A monster was killed
// MD   PlayerNum   DamageType  MonsterClass
function MonsterDeath(PlayerReplicationInfo Killer, Pawn Victim, class<DamageType> Damage)
{
    local string out;

    if (Killer == none)
        return;     //The monster suicided

    if (Killer != none && Killer.bBot)
    {
        if (!Checkbot(Killer))
            return;
    } else if (Killer != none && !Killer.bBot)
        CheckPlayer(Killer);

    out = Header()$"MD"$Tab;

    // KillerNumber and KillerDamagetype

    out $= Controller(Killer.Owner).PlayerNum$Tab;
    // KillerWeapon no longer used, using damagetype
    out $= GetItemName(string(Damage))$Tab;

    out $= GetItemName(string(Victim));

    Logf(out);

    ScoreEvent(Killer, Monster(Victim).ScoringValue, "invasion_frag");
    TeamScoreEvent(Killer.Team.TeamIndex, Monster(Victim).ScoringValue, "invasion_frag");
}

// A monster killed a player
// MK   PlayerNum   MonsterClass
function MonsterKill(Pawn Killer, PlayerReplicationInfo Victim)
{
    local string out, vehiclename;

    if (Killer == none)
        return;     //This should never happen

    if (Victim != none && Victim.bBot)
    {
        if (!Checkbot(Victim))
            return;
    } else if (Victim != none && !Victim.bBot)
        CheckPlayer(Victim);

    out = Header()$"MK"$Tab;

    // VictimNumber
    out $= Controller(Victim.Owner).PlayerNum$Tab;
    // Class of Killer
    out $= GetItemName(string(Killer))$Tab;

    // VictimNumber and VictimWeapon
    if ( bLogVehicleDeaths && Vehicle(Controller(Victim.Owner).Pawn) != none ) // Player is in a vehicle
    {
        vehiclename = GetItemName(string(Controller(Victim.Owner).Pawn));
            while( Asc(right(vehiclename,1)) >= 48 && Asc(right(vehiclename,1)) <= 57 ) // Strip off any numbers at the end of the name
                vehiclename = left(vehiclename,len(vehiclename)-1);
        out $= vehiclename;
    }
    else
        out $= GetItemName(string( GetLastWeapon(Controller(Victim.Owner)) ));

    Logf(out);

    // Invasion changes the scores but never calls the events for the log, so here they are
    ScoreEvent(Victim, -10, "invasion_death");
    TeamScoreEvent(Victim.Team.TeamIndex, -10, "invasion_death");
}

function ServerInfo()
{
    local string out, flags;
    local string siServerName, siAdminName, siAdminEmail;
    local GameInfo.ServerResponseLine ServerState;
    local int i;

    siServerName    = GRI.ServerName;   // Making local copies
    siAdminName     = GRI.AdminName;
    siAdminEmail    = GRI.AdminEmail;
    ReplaceText(siServerName,   tab, "_");
    ReplaceText(siAdminName,    tab, "_");
    ReplaceText(siAdminEmail,   tab, "_");

    out = Header() $ "SI" $ Tab;    // "SeverInfo"
    out $= siServerName $ Tab;      // Server name
    out $= TimeZone()$Tab;          // Timezone
    out $= siAdminName$Tab;         // Admin name
    out $= siAdminEmail$Tab;        // Admin email
    out $= Tab;                     // IP:port (filled in by Master Server)

    flags = "";                         // Server flag / key combos
    Level.Game.GetServerDetails( ServerState );
    for( i=0;i<ServerState.ServerInfo.Length;i++ )
        flags $= "\\"$ServerState.ServerInfo[i].Key$"\\"$ServerState.ServerInfo[i].Value;

    flags $= "\\gamedifficulty\\"$Level.Game.GameDifficulty;
    flags $= "\\shortname\\"$Level.Game.GameReplicationInfo.ShortName;

    ReplaceText(flags, tab, "_");
    out $= flags;

    Logf(out);

//    if(bExportHTMLLogs)
//        EGS.ServerInfo();
}

function StartGame()
{
    Logf( Header()$"SG" );          // "StartGame"


    // SPECIAL INVASION HANDLING
    if(Invasion(Level.Game) != none)
    {
        InvasionWaveCache = Invasion(Level.Game).InitialWave;
        GameEvent("new_invasion_wave", string(InvasionWaveCache), none);
    }

    if(bExportHTMLLogs)
        EGS.StartMatch();
}


// Send stats for the end of the game
function EndGame(string Reason)
{
    local string out;
    local int i,j;
    local array<PlayerReplicationInfo> PRIs;
    local PlayerReplicationInfo PRI,t;
    local Controller C;

    if(bEndGameExecuted)
        return;

    bEndGameExecuted = true;
    EndGameReason = Reason;

    out = Header()$"EG"$Tab$Reason; // "EndGame"

    // Quick cascade sort.
    for (i=0;i<GRI.PRIArray.Length;i++)
    {
        PRI = GRI.PRIArray[i];
        if ( !PRI.bOnlySpectator && (!PRI.bBot || bLogBots) )
        {
            PRIs.Length = PRIs.Length+1;
            for (j=0;j<Pris.Length-1;j++)
            {
                if (PRIs[j].Score < PRI.Score ||
                   (PRIs[j].Score == PRI.Score && PRIs[j].Deaths > PRI.Deaths) )
                {
                    t = PRIs[j];
                    PRIs[j] = PRI;
                    PRI = t;
                }
            }
            PRIs[j] = PRI;
        }
    }

    // Minimal scoreboard, shows Playernumbers in order of Score
    for (i=0;i<PRIs.Length;i++)
        out $= Tab$Controller(PRIs[i].Owner).PlayerNum;

    Logf(out);

    for( C=Level.ControllerList;C!=None;C=C.NextController )
    {
        if (C.PlayerReplicationInfo != none)
            ReportAccuracy(C.PlayerReplicationInfo);
    }

    // If user has Upload Logs turned on, do it now.
    if (bUploadLogs)
    {
        SendLogToServer();
    }


    // This makes it so the mean gameinfo can't deletes us!
    // It causes an accessed nones but we don't cares!
    // This allows chat to be logged after the match is over.
    Level.Game.Gamestats = none;
    bGameStatsJustRemoved = true;

    if (bExportHTMLLogs && !bExportedHTMLLogs)
    {
        if(Reason ~= "LastMan" || Reason ~= "ScoreLimit" || Reason ~= "TimeLimit")
        {
            // Wait until next tick to export the stats, that way the game lags
            // AFTER the camera is switched to the winner and the announcements play.
        }
        else
        {
            // This was probably a serverquit or a mapchange, so export the stats now.
            EGS.EndGame(EndGameReason, TheLogName);
            bExportedHTMLLogs = true;
        }
    }
}


// Connect Events get fired every time a player connects to a server
function ConnectEvent(PlayerReplicationInfo Who)
{
    local string out;
    local OLStatsAccuracyCounter AC;

    if ( Who.bOnlySpectator )           // Spectators should never show up in stats!
        return;

    if (Who.bBot)
    {
        Checkbot(Who);
        return;
    }

    // C    0   11d8944d9e138a5aa688d503e0e4c3e0
    out = Header()$"C"$Tab$Controller(Who.Owner).PlayerNum$Tab;

    // Login identifier
    out $= GetStatsIdentifier(Controller(Who.Owner));

    Logf(out);

    if (bExportHTMLLogs)
        EGS.PlayerConnected(Who, false);

    if (PlayerController(Who.Owner) != none && PlayerController(Who.Owner).Player != none && bLogSecurityInfo)
    {
        out = Header()$"PS"$Tab$Controller(Who.Owner).PlayerNum$Tab;
        // Player IP and netspeed
        out $= PlayerController(Who.Owner).GetPlayerNetworkAddress()$Tab$PlayerController(Who.Owner).Player.CurrentNetSpeed$Tab;
        // Player ID hash, based on CD key.
        out $= PlayerController(Who.Owner).GetPlayerIDHash();
        Logf(out);
    }

    if ( xLastManStandingGame(Level.Game) != none )
    {
        out = Header()$"G"$Tab$"lives"$Tab$Controller(Who.Owner).PlayerNum$Tab$Level.Game.MaxLives;
        Logf(out);
    }

    if(bLogAccuracy)
    {
        // Spawn the accuracy counter for this player
        AC = spawn(class'OLStatsAccuracyCounter');
        AC.PRI = Who;
        AC.GS = self;
        AddAccuracyCounter(AC);
    }

    PlayerArray[PlayerArray.length] = Controller(Who.Owner).PlayerNum;
}

// Connect Events get fired every time a player connects or leaves from a server
function DisconnectEvent(PlayerReplicationInfo Who)
{
    local string out;
    local int i;

    if ( Who.bOnlySpectator )           // Spectators should never show up in stats!
        return;

    if (Who.bBot)
    {
        if (!Checkbot(Who))
            return;

        // Remove the bot from the bot array
        for (i=0;i<BotArray.length;i++)
        {
            if(Controller(Who.Owner).PlayerNum == BotArray[i])
            {
                BotArray.Remove(i,1);
                break;
            }
        }
    } else {
        CheckPlayer(Who);
        // Remove the player from the player array
        for (i=0;i<PlayerArray.length;i++)
        {
            if(Controller(Who.Owner).PlayerNum == PlayerArray[i])
            {
                PlayerArray.Remove(i,1);
                break;
            }
        }
    }

    // D    0
    out = Header()$"D"$Tab$Controller(Who.Owner).PlayerNum; //"Disconnect"

    Logf(out);

    if (bExportHTMLLogs)
        EGS.Disconnect(Who);

    ReportAccuracy(Who);
}

function ReportAccuracy(PlayerReplicationInfo Who)
{
    local string out;
    local OLStatsAccuracyCounter AccuracyCounter;
    local int i;

    AccuracyCounter = GetAccuracyCounter(Who);

    if(!bLogAccuracy || Controller(Who.Owner) == none || AccuracyCounter == none)
        return;

    for(i=0;i<AccuracyCounter.WeaponAccuracyArray.Length;i++)
    {
        // If a player fired 0 shots, then skip this weapon.
        if (AccuracyCounter.WeaponAccuracyArray[i].ShotsFired == 0)
            continue;
        out = Header()$"PA"$Tab$Controller(Who.Owner).PlayerNum$Tab;
        out $= AccuracyCounter.WeaponAccuracyArray[i].WeaponName$Tab;
        out $= AccuracyCounter.WeaponAccuracyArray[i].ShotsFired$Tab;
        out $= AccuracyCounter.WeaponAccuracyArray[i].ShotsHit$Tab;
        out $= AccuracyCounter.WeaponAccuracyArray[i].TotalDamage;
        Logf(out);
    }

    RemoveAccuracyCounter(Who);

}


// Scoring Events occur when a player's score changes
function ScoreEvent(PlayerReplicationInfo Who, float Points, string Desc)
{
    if ( Who.bOnlySpectator )           // Just to be totally safe Spectators sends nothing
        return;

    if (Who.bBot)
    {
        if (!Checkbot(Who))
            return;
    } else
        CheckPlayer(Who);

    // SPECIAL MUTANT HANDLING
    // Mutant reports the scores all wrong, so I override the score event and
    // handle it myself from OLGameStatRules
    if ( xMutantGame(Level.Game) != none )
    {
        return;
    }

    Logf( Header()$"S"$Tab$Controller(Who.Owner).PlayerNum$Tab$Points$Tab$Desc );   //"Score"

    if (bExportHTMLLogs)
        EGS.ScoreEvent(Who, Points, Desc);
}

// Same as a score event, but it doesn't check for Mutant validity.
function MutantScoreEvent(PlayerReplicationInfo Who, float Points, string Desc)
{
    if ( Who.bOnlySpectator )           // Just to be totally safe Spectators sends nothing
        return;

    if (Who.bBot)
    {
        if (!Checkbot(Who))
            return;
    } else
        CheckPlayer(Who);


    Logf( Header()$"S"$Tab$Controller(Who.Owner).PlayerNum$Tab$Points$Tab$Desc );   //"Score"

    if (bExportHTMLLogs)
        EGS.ScoreEvent(Who, Points, Desc);
}


function TeamScoreEvent(int Team, float Points, string Desc)
{
    // SPECIAL HANDLING FOR INVASION
    // The team score event is called manually by me when a monster dies so the
    // right number of points are reported.
    if ( Invasion(Level.game) != none && Desc == "tdm_frag" )
        return;

    Logf( Header()$"T"$Tab$Team$Tab$Points$Tab$Desc );  //"TeamScore"

    if (bExportHTMLLogs)
        EGS.TeamScoreEvent(Team,Points,Desc);
}


// KillEvents occur when a player kills, is killed, suicides
function KillEvent(string Killtype, PlayerReplicationInfo Killer, PlayerReplicationInfo Victim, class<DamageType> Damage)
{
    local string out, vehiclename, damagename;
    local string HeldWeapon;

    if ( Victim.bOnlySpectator )
        return;

    if (Victim.bBot)
    {
        if (!Checkbot(Victim))
            return;
    } else
        CheckPlayer(Victim);

    if (Killer != none && Killer.bBot)
    {
        if (!Checkbot(Killer))
            return;
    } else if (Killer != none && !Killer.bBot)
        CheckPlayer(Killer);

    // SPECIAL INVASION HANDLING
    // If the game is invasion, return. Invasion kill events have their own function
    if ( Invasion(Level.Game) != none )
    {
        return;
    }

    out = Header()$Killtype$Tab;

    // KillerNumber and KillerDamagetype
    if (Killer!=None)
    {
        out $= Controller(Killer.Owner).PlayerNum$Tab;
        // KillerWeapon no longer used, using damagetype

        // FIX for the Epic bug in which the primary fire in the Leviathan uses the wrong DamageType
        if( GetItemName(string(Damage)) == "DamTypeRocket" && ONSMobileAssaultStation(Controller(Killer.Owner).Pawn) != none )
            Damagename = "DamTypeMASRocket";
        // FIX for the Epic bugs in which the Artillery's Side Turret uses the same damagetypes as the Hellbender
        else if( GetItemName(string(Damage)) == "DamTypePRVLaser" && Controller(Killer.Owner).Pawn.IsA('ONSArtillerySideGunPawn') )
            Damagename = "DamTypeArtilleryLaser";
        else if( GetItemName(string(Damage)) == "DamTypePRVCombo" && Controller(Killer.Owner).Pawn.IsA('ONSArtillerySideGunPawn') )
            Damagename = "DamTypeArtilleryCombo";
        // FIX for the Epic bugs where none of the new ECE vehicles have their own roadkill or pancake damtypes
        else if( GetItemName(string(Damage)) == "DamTypeRVRoadkill" && Controller(Killer.Owner).Pawn.IsA('ONSShockTank') )
            Damagename = "DamTypeShockTankRoadkill";
        else if( GetItemName(string(Damage)) == "DamTypeRVPancake" && Controller(Killer.Owner).Pawn.IsA('ONSShockTank') )
            Damagename = "DamTypeShockTankPancake";
        else if( GetItemName(string(Damage)) == "DamTypePRVRoadkill" && Controller(Killer.Owner).Pawn.IsA('ONSArtillery') )
            Damagename = "DamTypeArtilleryRoadkill";
        else if( GetItemName(string(Damage)) == "DamTypePRVPancake" && Controller(Killer.Owner).Pawn.IsA('ONSArtillery') )
            Damagename = "DamTypeArtilleryPancake";
        else if( GetItemName(string(Damage)) == "DamTypeAttackCraftRoadkill" && Controller(Killer.Owner).Pawn.IsA('ONSDualAttackCraft') )
            Damagename = "DamTypeDualAttackCraftRoadkill";
        else if( GetItemName(string(Damage)) == "DamTypeAttackCraftPancake" && Controller(Killer.Owner).Pawn.IsA('ONSDualAttackCraft') )
            Damagename = "DamTypeDualAttackCraftPancake";
        else
            Damagename = GetItemName(string(Damage));

        out $= Damagename$Tab;
    }
    else
    {
        Damagename = GetItemName(string(Damage));
        out $= "-1"$Tab$DamageName$Tab;    // No PlayerNum -> -1, Environment "deaths"
    }

    // VictimNumber and VictimWeapon
    if ( bLogVehicleDeaths && Vehicle(Controller(Victim.Owner).Pawn) != none ) // Player is in a vehicle
    {
        vehiclename = GetItemName(string(Controller(Victim.Owner).Pawn));
            while( Asc(right(vehiclename,1)) >= 48 && Asc(right(vehiclename,1)) <= 57 ) // Strip off any numbers at the end of the name
                vehiclename = left(vehiclename,len(vehiclename)-1);
        out $= Controller(Victim.Owner).PlayerNum$Tab$vehiclename;
        HeldWeapon = vehiclename;
    }
    else
    {
        out $= Controller(Victim.Owner).PlayerNum$Tab$GetItemName(string( GetLastWeapon(Controller(Victim.Owner)) ));
        HeldWeapon = GetItemName(string( GetLastWeapon(Controller(Victim.Owner)) ));
    }

    // Type killers tracked as player event (redundant Typing, removed from kill line)
    if ( PlayerController(Victim.Owner)!= None && PlayerController(Victim.Owner).bIsTyping)
    {
        if ( PlayerController(Killer.Owner) != PlayerController(Victim.Owner) )
            SpecialEvent(Killer, "type_kill");                      // Killer killed typing victim
    }

    Logf(out);

    if(bExportHTMLLogs)
        EGS.PlayerKilledPlayer(Killer,Victim,DamageName,HeldWeapon);
}


// Special Invasion KillEvent.
// KillEvents occur when a player kills, is killed, suicides
function InvasionKillEvent(PlayerReplicationInfo Killer, PlayerReplicationInfo Victim, class<DamageType> Damage)
{
    local string out, vehiclename, damagename, Killtype;
    local string HeldWeapon;

    if ( Victim.bOnlySpectator )
        return;

    if (Victim.bBot)
    {
        if (!Checkbot(Victim))
            return;
    } else
        CheckPlayer(Victim);

    if (Killer != none && Killer.bBot)
    {
        if (!Checkbot(Killer))
            return;
    } else if (Killer != none && !Killer.bBot)
    CheckPlayer(Killer);

    if (Killer != none && Killer != Victim && Killer.Team.TeamIndex == Victim.Team.TeamIndex)
        Killtype = "TK";
    else
        Killtype = "K";

    out = Header()$Killtype$Tab;

    // KillerNumber and KillerDamagetype
    if (Killer!=None)
    {
        out $= Controller(Killer.Owner).PlayerNum$Tab;
        // KillerWeapon no longer used, using damagetype

        // FIX for the Epic bug in which the primary fire in the Leviathan uses the wrong DamageType
        if( GetItemName(string(Damage)) == "DamTypeRocket" && ONSMobileAssaultStation(Controller(Killer.Owner).Pawn) != none )
            Damagename = "DamTypeMASRocket";
        // FIX for the Epic bugs in which the Artillery's Side Turret uses the same damagetypes as the Hellbender
        else if( GetItemName(string(Damage)) == "DamTypePRVLaser" && Controller(Killer.Owner).Pawn.IsA('ONSArtillerySideGunPawn') )
            Damagename = "DamTypeArtilleryLaser";
        else if( GetItemName(string(Damage)) == "DamTypePRVCombo" && Controller(Killer.Owner).Pawn.IsA('ONSArtillerySideGunPawn') )
            Damagename = "DamTypeArtilleryCombo";
        // FIX for the Epic bugs where none of the new ECE vehicles have their own roadkill or pancake damtypes
        else if( GetItemName(string(Damage)) == "DamTypeRVRoadkill" && Controller(Killer.Owner).Pawn.IsA('ONSShockTank') )
            Damagename = "DamTypeShockTankRoadkill";
        else if( GetItemName(string(Damage)) == "DamTypeRVPancake" && Controller(Killer.Owner).Pawn.IsA('ONSShockTank') )
            Damagename = "DamTypeShockTankPancake";
        else if( GetItemName(string(Damage)) == "DamTypePRVRoadkill" && Controller(Killer.Owner).Pawn.IsA('ONSArtillery') )
            Damagename = "DamTypeArtilleryRoadkill";
        else if( GetItemName(string(Damage)) == "DamTypePRVPancake" && Controller(Killer.Owner).Pawn.IsA('ONSArtillery') )
            Damagename = "DamTypeArtilleryPancake";
        else if( GetItemName(string(Damage)) == "DamTypeAttackCraftRoadkill" && Controller(Killer.Owner).Pawn.IsA('ONSDualAttackCraft') )
            Damagename = "DamTypeDualAttackCraftRoadkill";
        else if( GetItemName(string(Damage)) == "DamTypeAttackCraftPancake" && Controller(Killer.Owner).Pawn.IsA('ONSDualAttackCraft') )
            Damagename = "DamTypeDualAttackCraftPancake";
        else
            Damagename = GetItemName(string(Damage));

        out $= Damagename$Tab;
    }
    else
    {
        Damagename = GetItemName(string(Damage));
        out $= "-1"$Tab$DamageName$Tab;    // No PlayerNum -> -1, Environment "deaths"
    }

    // VictimNumber and VictimWeapon
    if ( bLogVehicleDeaths && Vehicle(Controller(Victim.Owner).Pawn) != none ) // Player is in a vehicle
    {
        vehiclename = GetItemName(string(Controller(Victim.Owner).Pawn));
            while( Asc(right(vehiclename,1)) >= 48 && Asc(right(vehiclename,1)) <= 57 ) // Strip off any numbers at the end of the name
                vehiclename = left(vehiclename,len(vehiclename)-1);
        out $= Controller(Victim.Owner).PlayerNum$Tab$vehiclename;
        HeldWeapon = vehiclename;
    }
    else
    {
        out $= Controller(Victim.Owner).PlayerNum$Tab$GetItemName(string( GetLastWeapon(Controller(Victim.Owner)) ));
        HeldWeapon = GetItemName(string( GetLastWeapon(Controller(Victim.Owner)) ));
    }

    // Type killers tracked as player event (redundant Typing, removed from kill line)
    if ( PlayerController(Victim.Owner)!= None && PlayerController(Victim.Owner).bIsTyping)
    {
        if ( PlayerController(Killer.Owner) != PlayerController(Victim.Owner) )
            SpecialEvent(Killer, "type_kill");                      // Killer killed typing victim
    }

    Logf(out);

    if(bExportHTMLLogs)
        EGS.PlayerKilledPlayer(Killer,Victim,DamageName,HeldWeapon);


    // SPECIAL INVASION SCORING

    ScoreEvent(Victim, -10, "invasion_death");
    TeamScoreEvent(Victim.Team.TeamIndex, -10, "invasion_death");

    if (Killer!= none && Killer != Victim && Killer.Team.TeamIndex == Victim.Team.TeamIndex)
        TeamScoreEvent(Killer.Team.TeamIndex, -10, "team_frag");
    else if (Killer == Victim)
    {
        ScoreEvent(Victim, -1, "self_frag");
        TeamScoreEvent(Victim.Team.TeamIndex, -1, "self_frag");
    }

}

function class<Weapon> GetLastWeapon(Controller Other)
{
    if (RPGHelper != none)
    {
        return RPGHelper.RPGGetLastWeapon(Other);
    }

    return Other.GetLastWeapon();
}


// Special Events are everything else regarding the player
function SpecialEvent(PlayerReplicationInfo Who, string Desc)
{
    local string out;
    if (Who != None)
    {
        if ( Who.bOnlySpectator )       // Avoid spectator suicide on console "type_kill"
            return;

        if (Who.bBot)
        {
            if (!Checkbot(Who))
                return;
        } else
            CheckPlayer(Who);

        out = string(Controller(Who.Owner).PlayerNum);
    }
    else
        out = "-1";

    Logf( Header()$"P"$Tab$out$Tab$Desc );                  //"PSpecial"

    if(bExportHTMLLogs)
        EGS.SpecialEvent(Who,Desc);
}


// Special events regarding the game
function GameEvent(string GEvent, string Desc, PlayerReplicationInfo Who)
{
    local string out, geDesc;

    if (Who != None)
    {
        if ( Who.bOnlySpectator )       // Specator could cause NameChange, TeamChange! No longer.
            return;

        if (Who.bBot)
        {
            if (!Checkbot(Who))
                return;
        } else
            CheckPlayer(Who);

        out = string(Controller(Who.Owner).PlayerNum);
    }
    else
        out = "-1";

    geDesc  = Desc;
    ReplaceText(geDesc, tab, "_");                                  // geDesc, can be the nickname!

    Logf( Header()$"G"$Tab$GEvent$Tab$out$Tab$geDesc ); //"GSpecial"

    if(bExportHTMLLogs)
        EGS.GameEvent(Who,GEvent,Desc);
}

function CheckPlayer(PlayerReplicationInfo Who)
{
    local int i;
    local bool bFound;

    for (i=0;i<PlayerArray.length;i++)
    {
        if( Controller(Who.Owner).PlayerNum == PlayerArray[i] )
            bFound = true;
    }

    if (bFound)
    {
        return;
    }
    else
    {
        ConnectEvent(Who);
        GameEvent("NameChange",Who.playername,Who);
        if (Who.Team != none && Who.Team.TeamIndex < 255)
        {
            GameEvent("TeamChange",string(Who.Team.TeamIndex),Who);
        }
    }
}

function bool Checkbot(PlayerReplicationInfo Who)
{
    local int i;
    local bool bFound;
    local string out;
    local OLStatsAccuracyCounter AC;
    local bot b;

    bFound = false;

    if (!bLogBots)
        return false;

    // Monsters should not get recorded as players - happens with UT2004RPG
    if ( MonsterController(Who.Owner) != none )
        return false;

    for (i=0;i<BotArray.length;i++)
    {
        if(Controller(Who.Owner).PlayerNum == BotArray[i])
        {
            bFound = true;
            continue;
        }
    }

    if (!bFound)
    {
        if(bExportHTMLLogs)
        {
            EGS.PlayerConnected(Who, true);
        }

        BotArray[BotArray.length] = Controller(Who.Owner).PlayerNum;
        out = header()$"C"$tab$Controller(Who.Owner).PlayerNum$tab$Who.playername;
        Logf(out);

        GameEvent("NameChange","[BOT]"$Who.playername,Who);
        if (Who.Team != none && Who.Team.TeamIndex < 255)
        {
            GameEvent("TeamChange",string(Who.Team.TeamIndex),Who);
        }

        // Bot skill reporting
        // [time] BI [playernum] [skill] [BaseAlertness] [Accuracy] [BaseAggressiveness] [StrafingAbility] [CombatStyle] [Tactics] [TranslocUse] [Jumpiness] [FavoriteWeapon]
        b = bot(Who.Owner);
        if (b != none)
        {
            out = header()$"BI"$tab$b.PlayerNum$tab$b.skill$tab$b.BaseAlertness$tab$b.Accuracy$tab$b.BaseAggressiveness$tab$b.StrafingAbility$tab$b.CombatStyle$tab$b.Tactics$tab$b.TranslocUse$tab$b.Jumpiness$tab$GetItemName(string(b.FavoriteWeapon));
            Logf(out);
        }

        if ( xLastManStandingGame(Level.Game) != none )
        {
            out = Header()$"G"$Tab$"lives"$Tab$Controller(Who.Owner).PlayerNum$Tab$Level.Game.MaxLives;
            Logf(out);
        }

        if(bLogAccuracy)
        {
            // Spawn the accuracy counter for this player
            AC = spawn(class'OLStatsAccuracyCounter');
            if (AC != none)
            {
                AC.PRI = Who;
                AC.GS = self;
                AddAccuracyCounter(AC);
            }
        }

    }

    return true;
}

function DamageDealt(PlayerReplicationInfo Who, class<DamageType> DamageType, int Damage)
{
    local OLStatsAccuracyCounter AC;

    if( !bLogAccuracy )
        return;

    AC = GetAccuracyCounter(Who);

    if (AC != none)
        AC.DamageDealt(DamageType, Damage);
}

function AddAccuracyCounter(OLStatsAccuracyCounter AC)
{
    AccuracyCounters[AccuracyCounters.Length] = AC;
}

function RemoveAccuracyCounter(PlayerReplicationInfo Who)
{
    local int Pointer;
    local OLStatsAccuracyCounter AC;

    AC = GetAccuracyCounter(Who, Pointer);

    if (AC != none)
    {
        AccuracyCounters.Remove(Pointer,1);
        AC.destroy();
    }
}

function OLStatsAccuracyCounter GetAccuracyCounter(PlayerReplicationInfo Who, optional out int Pointer)
{
    local int i;

    if (Who == none)
        return none;

    for(i=0;i<AccuracyCounters.Length;i++)
    {
        if(AccuracyCounters[i].PRI == Who)
        {
            Pointer = i;
            return AccuracyCounters[i];
        }
    }
    return none;
}

function Tick(float Delta)
{
    local GameInfo.ServerResponseLine SRL;
    local xDoubleDom DDomGame;
    local int i;
    local string out;

    TimeSinceLastPingReport += Delta;

    if(bGameStatsJustRemoved)
    {
        Level.Game.GameStats = self;
        if (bExportHTMLLogs && !bExportedHTMLLogs)
        {
           EGS.EndGame(EndGameReason, TheLogName);
           bExportedHTMLLogs = true;
        }
    }

    if(TimeSinceLastPingReport >= PingReportInterval && bLogPings)
    {
        TimeSinceLastPingReport=0;
        level.game.GetServerPlayers(SRL);
        for (i=0;i<SRL.PlayerInfo.length;i++)
        {
            out = Header()$"PP"$Tab$SRL.PlayerInfo[i].PlayerNum$Tab$SRL.PlayerInfo[i].Ping;
            Logf(out);
        }
    }

    // SPECIAL DOMINATION HANDLING
    // Checks to see what team controls each point.  If it's different, then it
    // reports a game event showing who captured the point.
    if (xDoubleDom(Level.game) != none)
    {
        DDomGame = xDoubleDom(Level.game);
        for(i=0;i<2;i++)
        {
            if(DDomGame.xDomPoints[i].bControllable && DDomGame.xDomPoints[i].ControllingTeam != none && DDomGame.xDomPoints[i].ControllingTeam != DomPointTeamCache[i])
            {
                DomPointTeamCache[i] = DDomGame.xDomPoints[i].ControllingTeam;
                GameEvent("dom_point_capture",string(i),DDomGame.xDomPoints[i].ControllingPawn.PlayerReplicationInfo);
            }
            else if (DomPointTeamCache[i] != none && (!DDomGame.xDomPoints[i].bControllable || DDomGame.xDomPoints[i].ControllingTeam == none))
            {
                DomPointTeamCache[i] = none;
                GameEvent("dom_point_capture",string(i),none);
            }
        }
    }

    // SPECIAL MUTANT HANDLING
    // Checks if there is a new mutant or bottom feeder
    if (xMutantGame(Level.game) != none)
    {
        if (xMutantGame(Level.game).CurrentMutant != none && xMutantGame(Level.game).CurrentMutant != MutantCache)
        {
            MutantCache = xMutantGame(Level.game).CurrentMutant;
            GameEvent("new_mutant", "", MutantCache.PlayerReplicationInfo);
        }
        else if (xMutantGame(Level.game).CurrentMutant == none && MutantCache != none)
        {
            MutantCache = none;
            GameEvent("new_mutant", "", none);
        }
        if (xMutantGame(Level.game).CurrentBottomFeeder != none && xMutantGame(Level.game).CurrentBottomFeeder != BottomFeederCache)
        {
            BottomFeederCache = xMutantGame(Level.game).CurrentBottomFeeder;
            GameEvent("new_bottomfeeder", "", BottomFeederCache.PlayerReplicationInfo);
        }
        else if (xMutantGame(Level.game).CurrentBottomFeeder == none && BottomFeederCache != none)
        {
            BottomFeederCache = none;
            GameEvent("new_bottomfeeder", "", none);
        }
    }

    // SPECIAL INVASION HANDLING
    // This detects if a new invasion wave has begun by checking to see if it's changed
    if (Invasion(level.game) != none)
    {
        if ( Invasion(level.game).WaveNum != InvasionWaveCache )
        {
            InvasionWaveCache = Invasion(level.game).WaveNum;
            GameEvent("new_invasion_wave", string(InvasionWaveCache), none);
        }
    }
}

/**
    return the filename to use for the log file. The following formatting rules are accepted:
    %P      server port
    %Y      year
    %M      month
    %D      day
    %H      hour
    %I      minute
    %S      second
    %W      day of the week
*/
function string GetLogFilename()
{
  local string result;
  result = LogFileName;
  ReplaceText(result, "%P", string(Level.Game.GetServerPort()));
  ReplaceText(result, "%N", Level.Game.GameReplicationInfo.ServerName);
  ReplaceText(result, "%Y", Right("0000"$string(Level.Year), 4));
  ReplaceText(result, "%M", Right("00"$string(Level.Month), 2));
  ReplaceText(result, "%D", Right("00"$string(Level.Day), 2));
  ReplaceText(result, "%H", Right("00"$string(Level.Hour), 2));
  ReplaceText(result, "%I", Right("00"$string(Level.Minute), 2));
  ReplaceText(result, "%W", Right("0"$string(Level.DayOfWeek), 1));
  ReplaceText(result, "%S", Right("00"$string(Level.Second), 2));
  return result;
}

defaultproperties
{
    bAlwaysTick=true
    LogFileName="Stats_%P_%Y_%M_%D_%H_%I_%S"
    Version="3.02"
    PingReportInterval=30
    bDebug=false
    RPGHelperName="OLStatsRPG.OLStatsRPGHelper"
    DateTimeFormat = "%e, %f %D %Y at %g:%I:%S %A"
    DateFormat = "%e, %f %D %Y"
}
